/*
  Hotsos Symposium 2008
  
  Presentation - "Workload Management with Oracle RAC"
  by Alex Gorbachev, The Pythian Group

  Demo 8 - Run-time Load Balancing example
*/

import java.sql.*;
import java.util.*;
import oracle.jdbc.pool.*;
import oracle.jdbc.*;

public class th extends Thread {

   int m_myId; // my own id (when it's a thread)

   static int m_nextId = 1; // next Id of the thread to start
   static OracleDataSource ods; // shared amongst all threads
   static int[] tps = new int[3]; // this array stores number of transactions

   synchronized static int getNextId()
   {
      return m_nextId++;
   }

   public static void main(String[] args) throws InterruptedException {
     try {
        int THREADS = 20;         // Number of parallel threads
        OracleConnectionCacheManager occm = null;
        // Starting as usual...
        ods = new OracleDataSource();
        ods.setUser("oracloid");
        ods.setPassword("me");
	// Set connection name so that we can retrieve it later
        String cacheName = "cache";
	// RAC URL
        String dbURL="jdbc:oracle:thin:@\n" +
	      "(DESCRIPTION=(FAILOVER=ON)(LOAD_BALANCE=ON)\n" +
              "(ADDRESS=(PROTOCOL=TCP)(HOST=u1-vip)(PORT=1521))\n" +
              "(ADDRESS=(PROTOCOL=TCP)(HOST=u2-vip)(PORT=1521))\n" +
              "(ADDRESS=(PROTOCOL=TCP)(HOST=u3-vip)(PORT=1521))\n" +
              "(CONNECT_DATA=(SERVICE_NAME=s10gr2)))";
        System.out.println("Url = " + dbURL );
        ods.setURL(dbURL);
	// Enable Implicit Connection Cache
        ods.setConnectionCachingEnabled(true);
	// Set properties
        Properties prop = new Properties();
        prop.setProperty("MinLimit", "30");     // initial/max/minimal number of connections in the pool
        prop.setProperty("MaxLimit", "30");     // we are overallocating it for use with CLB_GOAL_LONG
        prop.setProperty("InitialLimit", "30");
	// Special retry mechanism in JDBC - to detect failure earlier
	// Metalink Note 453347.1
        prop.put (oracle.net.ns.SQLnetDef.TCP_CONNTIMEOUT_STR,"" + (1000)); // 1000 ms
        ods.setConnectionCacheProperties(prop);
	// Enable FCF causes RLB on
        ods.setFastConnectionFailoverEnabled(true);
	// Subscribe to remote ONS
        ods.setONSConfiguration("nodes=u1:6251,u2:6251");
	// We have only two nodes - avoid waiting on TCP timeout for u3
        //ods.setONSConfiguration("nodes=u1:6251,u2:6251,u3:6251");
        ods.setConnectionCacheName(cacheName);
        occm = OracleConnectionCacheManager.getConnectionCacheManagerInstance();
        // Initial connections are allocated on the first request
        // if it fails - I want to see it right away here
        Connection tmp = ods.getConnection();
        tmp.close();
        
        // Create the threads
        Thread[] threadList = new Thread[THREADS];
        // spawn threads
        for (int i = 0; i < THREADS; i++)
        {
          threadList[i] = new th();
          threadList[i].start();
          Thread.sleep(100);
        }
        
        for (int i = 0; i < THREADS; i++)
        {
          // Wait for every thread to finish
          while ( threadList[i].isAlive() ) {
            threadList[i].join(5000);
            int[] t = new int[3];
            synchronized (tps) {
              t[0]=tps[0]/5; tps[0]=0;
              t[1]=tps[1]/5; tps[1]=0;
              t[2]=tps[2]/5; tps[2]=0;
            }
            System.out.println("TPS per instance:\t"+t[0]+"\t"+t[1]+"\t"+t[2]+
                               "\tActive connections:\t" + 
			       occm.getNumberOfActiveConnections(cacheName) * 100 / 
			       (occm.getNumberOfActiveConnections(cacheName) +
			       occm.getNumberOfAvailableConnections(cacheName)) + '%');
          }  
        }
     }
     catch(Exception e)
     {
         e.printStackTrace();
     }
   }
   
   public th () {
     super();
     m_myId = getNextId();
   }
   
   public void run() {
     try {

        ResultSet rs = null;
        while (true){
          Connection conn = ods.getConnection();
          while ( conn == null ) {
            System.out.println("               Thread: " + m_myId + " NO MORE CONNECTIONS IN THE POOL! Waiting...");
            Thread.sleep(1000);
            conn = ods.getConnection();
          }
          Statement stmt=conn.createStatement();
          try {
            //rs = stmt.executeQuery("select instance_name from v$instance");
            //while(rs.next()) {
            //  System.out.println("               Thread: " + m_myId + " Instance name: " + rs.getString(1));
            stmt.execute("BEGIN NEWORDER; END;");
            conn.commit();
            ResultSet rset = stmt.executeQuery("select instance_number-1 from v$instance");
            rset.next();
            int inst = rset.getInt(1);
            synchronized (tps) {
              tps[inst]++;
            }
          }
          catch (SQLException sqle) {
            sqle.printStackTrace();
            System.out.print("               Thread: " + m_myId);
            System.out.print(" ErrorCode=" + sqle.getErrorCode());
            System.out.print(" SQLState=" + sqle.getSQLState());
            System.out.println(" Message=" + sqle.getMessage());
          }
          conn.close();
	  // delay between transactions
          Thread.sleep(500);
        }
 
     }
     catch (Exception e)
     {
       System.out.println("!!!!!!!!!!!!!! Thread: " + m_myId + " got Exception " + e + "!!!!!!!!!!!!!!");
       e.printStackTrace();
       return;
     }
   }
   
}
